/* FILE: mmdinit.cc                 -*-Mode: c++-*-
 *
 * Commands added to Tcl/Tk interpreter (& supporting routines)
 *
 */

#include <math.h>	/* fmod, ... */
#include <string.h>
#include <stdlib.h>
#include <tk.h>		/* Tk_Window, ... */

#include "nb.h"
#include "oc.h"
#include "vf.h"

#include "mmdisp.h"	/* MMDISP_MAJOR_VERSION, ... */
#include "mmdinit.h"	/* protoypes */
#include "display.h"	/* DisplayFrame ... */

/* End includes */     // This is an optional directive to build.tcl,
                       // that excludes the remainder of the file from
                       // '#include ".*"' dependency building

static char *ExeName=NULL;
static char *FirstFileName=NULL;

static Vf_Mesh *myMesh;
static DisplayFrame myFrame;

static INT4m ServerFlag=0;

#define MY_BUF_SIZE 4096
static CHAR buf[MY_BUF_SIZE];
static CHAR buf2[MY_BUF_SIZE];

void Mmd_SaveServerFlag(INT2m value)
{
  // Save the value in a static variable
  ServerFlag = value;
}

int TclCommandsPreInit(char* runname,char* inputfile)
{
  // Initialize myMesh with an empty mesh
  myMesh=new Vf_EmptyMesh();

  // Save executable name
  ExeName=new char[strlen(runname)+1];
  strcpy(ExeName,runname);
  // Save first image file name
  if(inputfile!=NULL) {
    FirstFileName=new char[strlen(inputfile)+1];
    strcpy(FirstFileName,inputfile);
  }
  return 0;
}

void TclCommandsCleanup(ClientData)
{
  // Free C++ allocated memory
  if(ExeName!=NULL) delete[] ExeName;  // Safety-check on Null
  if(FirstFileName!=NULL) delete[] FirstFileName;
  if(myMesh!=NULL) delete myMesh;
}

static REAL4m CoordsToAngle(CoordinateSystem coords)
{
  REAL4m rotang=0.;
  switch(coords) {
  case CS_CALCULATION_STANDARD: // Safety
  case CS_DISPLAY_STANDARD:  rotang=0.;   break;
  case CS_DISPLAY_ROT_90:    rotang=90.;  break;
  case CS_DISPLAY_ROT_180:   rotang=180.; break;
  case CS_DISPLAY_ROT_270:   rotang=270.; break;
  case CS_ILLEGAL:
  default:
    PlainError(1,"Unrecognized or illegal coordinate system (%d)"
	       " detected in CoordsToAngle() (File tkcmds.cc)\n",
	       int(coords));
    break; // Shouldn't get to here
  }
  return rotang;
}

static CoordinateSystem AngleToCoords(REAL4m angle_in_degrees)
{ // Rounds to nearest supported coordinate system rotation
  CoordinateSystem coords=CS_ILLEGAL;
  int quad=int(OC_ROUND(fmod(angle_in_degrees,360.)/90.));
  if(quad<0) quad+=4;  if(quad>3) quad-=4;
  switch(quad) {
  case 0:  coords=CS_DISPLAY_STANDARD;  break;
  case 1:  coords=CS_DISPLAY_ROT_90;    break;
  case 2:  coords=CS_DISPLAY_ROT_180;   break;
  case 3:  coords=CS_DISPLAY_ROT_270;   break;
  default:
    PlainError(1,"Programming error: Illegal coordinate system (quad=%d)"
	       " occurred in AngleToCoords() (File tkcmds.cc)\n",quad);
    break; // Shouldn't get to here
  }
  return coords;
}

int GetFirstFileName(ClientData,Tcl_Interp *interp,int argc,char **)
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    sprintf(buf,"GetFirstFileName must be called with no arguments"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  if(FirstFileName!=NULL) strncpy(buf,FirstFileName,sizeof(buf));
  else                    buf[0]='\0';
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int GetMeshType(ClientData,Tcl_Interp *interp,int argc,char**)
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    sprintf(buf,"GetMeshType must be called with no arguments"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  const char *name=myMesh->GetMeshType();
  if(name!=NULL) strncpy(buf,name,sizeof(buf));
  else           buf[0]='\0';
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int GetVecColor(ClientData,Tcl_Interp *interp,int argc,char** argv)
{
  Tcl_ResetResult(interp);
  if(argc!=2) {
    sprintf(buf,"SetMeshTitle must be called with 1 argument:"
	    " A 3 element list representing a vector"
            " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  int carr_size;
  char** carr;
  if (Tcl_SplitList(interp,argv[1],&carr_size,&carr)!=TCL_OK) {
    return TCL_ERROR;
  }
  if(carr_size!=3) {
    sprintf(buf,"Input list has %d != 3 elements",carr_size);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  Nb_Vec3<REAL4> v(atof(carr[0]),atof(carr[1]),atof(carr[2]));
  Tcl_Free((char *)carr);
  myFrame.GetVecColor(v);
  Oc_AutoBuf ab;
  Tcl_AppendResult(interp,ab(myFrame.GetVecColor(v)),(char *)NULL);
  return TCL_OK;
}


int GetMeshName(ClientData,Tcl_Interp *interp,int argc,char**)
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    sprintf(buf,"GetMeshName must be called with no arguments"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  const char *name=myMesh->GetName();
  if(name!=NULL) strncpy(buf,name,sizeof(buf));
  else           buf[0]='\0';
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int GetMeshTitle(ClientData,Tcl_Interp *interp,int argc,char**)
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    sprintf(buf,"GetMeshTitle must be called with no arguments"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  const char *title=myMesh->GetTitle();
  if(title!=NULL) strncpy(buf,title,sizeof(buf));
  else            buf[0]='\0';
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int SetMeshTitle(ClientData,Tcl_Interp *interp,int argc,char** argv)
{
  Tcl_ResetResult(interp);
  if(argc!=2) {
    sprintf(buf,"SetMeshTitle must be called with 1 argument:"
	    " new_title (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  myMesh->SetTitle(argv[1]);
  return TCL_OK;
}

int GetMeshDescription(ClientData,Tcl_Interp *interp,int argc,char**)
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    sprintf(buf,"GetMeshDesc must be called with no arguments"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  const char *desc=myMesh->GetDescription();
  if(desc!=NULL) strncpy(buf,desc,sizeof(buf));
  else           buf[0]='\0';
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int GetMeshIncrement(ClientData,Tcl_Interp *interp,int argc,char**)
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    sprintf(buf,"GetMeshIncrement must be called with no arguments"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  sprintf(buf,"%f",myMesh->GetSubsampleGrit());
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int ChangeMesh(ClientData cd,Tcl_Interp *interp,int argc,char** argv)
{
  Tcl_ResetResult(interp);
  if(argc<5 || argc>6) {
    sprintf(buf,"ChangeMesh must be called with 4 or 5 arguments:"
	    " mesh_filename"
	    " frame_width frame_height rotation_degrees"
	    " [zoom]"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  delete myMesh;
  if(argv[1]==NULL || argv[1][0]=='\0') { // Null request
    myMesh=new Vf_EmptyMesh();
  }
  else {
    Vf_FileInput *vffreader
      = Vf_FileInput::NewReader(argv[1]);
    if(vffreader==NULL) { // Unknown file type
      myMesh=new Vf_EmptyMesh();
    }
    else { // Reader found for file argv[1]
      myMesh=vffreader->NewMesh();
      delete vffreader;
      if(myMesh==NULL) myMesh=new Vf_EmptyMesh(); // Safety
    }
  }

  myFrame.SetMesh(myMesh);
  REAL4m width=Nb_Atof(argv[2]);
  REAL4m height=Nb_Atof(argv[3]);
  CoordinateSystem coords=AngleToCoords(Nb_Atof(argv[4]));
  REAL4m zoom=-1.;
  if(argc>5) {
    zoom=Nb_Atof(argv[5]);
  }

  // Adjust sizes to leave room for margins.
  Oc_AutoBuf ab1,ab2;
  char *cptr=Tcl_GetVar2(interp,ab1("plot_config"),
			 ab2("misc,margin"),TCL_GLOBAL_ONLY);
  if(cptr!=NULL && *cptr!='\0') {
    REAL4m margin=Nb_Atof(cptr);
    width-=2*margin;
    height-=2*margin;
  }

  myFrame.SetCoordinates(coords);

  if(zoom>0.) myFrame.SetZoom(zoom);
  else        myFrame.SetZoom(width,height);

  UpdatePlotConfiguration(cd,interp,0,NULL);
  /// For now, just use current configuration /**/

  return TCL_OK;
}

int DrawFrame(ClientData,Tcl_Interp *interp,int argc,char** argv)
{
  int error_code;
  Tcl_ResetResult(interp);
  if(argc!=2) {
    sprintf(buf,"DrawFrame must be called with 1 argument: canvas_name"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  sprintf(buf,"%s delete all ;# update idletasks",argv[1]);
  /// Uncomment the "update idletasks" command to get a white
  /// screen between frames
  Tcl_Eval(interp,buf); // Clear canvas
  error_code=myFrame.Show(argv[1]); // Redraw canvas

  sprintf(buf,"global DrawFrameCount; incr DrawFrameCount");
  Tcl_Eval(interp,buf); // Increment DrawFrameCount variable

/* Adjust subsample setting to actual value.
 * Note: Any tracings on plot_config triggered as a result
 *  of the Tcl_SetVar2 calls will send the name used in the
 *  call (here, e.g., "plot_config arrow_subsample") to the
 *  trace callback, and the call will appear to be from the
 *  proc that called this routine (DrawFrame).  If that routine
 *  has not made plot_config visible through a 'global plot_config'
 *  command, then a "upvar plot_config" will *fail* in the trace
 *  callback.  To protect against this, trace callbacks on plot_config
 *  should have the global name hardcoded into the trace call, and the
 *  callback itself should use the hardcoded value instead of the
 *  "name" argument.  For example:
 *
 *            trace variable plot_config(arrow_subsample) w \
 *                "MyCallback plot_config"
 *  and
 *        proc MyCallback { globalname name elt op } {
 *                upvar #0 $globalname x
 *                ...
 *        }
 *
 *  (The "elt" value can be either hardcoded into globalname, or
 *  processed separately.)
 */
  Oc_AutoBuf ab1,ab2;
  REAL4m arrow_sample,pixel_sample;
  REAL4m arrow_sample_request,pixel_sample_request;
  myFrame.GetRequestedSubsampleRates(arrow_sample_request,
				     pixel_sample_request);
  myFrame.GetActualSubsampleRates(arrow_sample,pixel_sample);
  if(arrow_sample_request<0.) {
    arrow_sample*=-1;
    if(arrow_sample > -0.01) arrow_sample = -0.01;
  }
  sprintf(buf,"%g",arrow_sample);
  Tcl_SetVar2(interp,ab1("plot_config"),ab2("arrow,subsample"),
	      buf,TCL_GLOBAL_ONLY);
  if(pixel_sample_request<0.) {
    pixel_sample*=-1;
    if(pixel_sample > -0.01) pixel_sample = -0.01;
  }
  sprintf(buf,"%g",pixel_sample);
  Tcl_SetVar2(interp,ab1("plot_config"),ab2("pixel,subsample"),
	      buf,TCL_GLOBAL_ONLY);

  if(error_code!=TCL_OK) {
    sprintf(buf,"Error from myFrame.Show(%s)",argv[1]);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return error_code;
  }
  return TCL_OK;
}

int GetFrameRotation(ClientData,Tcl_Interp *interp,
		     int argc,char** argv)
{
  Tcl_ResetResult(interp);
  if(argc!=2) {
    sprintf(buf,"GetFrameRotation must be called with"
	    " 1 argument: name of Tk's Frame Rotation variable"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  INT4m rotang=int(OC_ROUND(CoordsToAngle(myFrame.GetCoordinates())));

  sprintf(buf,"%d",rotang);
  Tcl_SetVar(interp,argv[1],buf,0);

  return TCL_OK;
}

// Following routine added by SLW May-1998, modified by MJD June-1998.
// It copies the plot configuration values from the Tcl variable
// plot_config into the corresponding Frame plot_config structures.
// Also sets quantitylist value from Frame.
int UpdatePlotConfiguration(ClientData,Tcl_Interp *interp,int,char **)
{
  Tcl_ResetResult(interp);
  // Ignore argument list; just read plot_config() directly and update
  // all fields.
  char *cptr; int colorcount;
  Oc_AutoBuf ab1,ab2;
  PlotConfiguration arrow_config,pixel_config;
  myFrame.GetPlotConfiguration(arrow_config,pixel_config);

  // Arrow plot configuration
  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("arrow,status"),
		   TCL_GLOBAL_ONLY);
  if(cptr!=NULL && *cptr!='\0')
    arrow_config.displaystate=atoi(cptr);

  colorcount=-1;
  cptr=Tcl_GetVar2(interp,ab1("plot_config"),
		   ab2("arrow,colorcount"),TCL_GLOBAL_ONLY);
  if(cptr!=NULL && *cptr!='\0') colorcount=atoi(cptr);
  if(colorcount<0) colorcount=arrow_config.colormap.GetColorCount();
  cptr=Tcl_GetVar2(interp,ab1("plot_config"),
			 ab2("arrow,colormap"),TCL_GLOBAL_ONLY);
  if(cptr!=NULL && *cptr!='\0')
    arrow_config.colormap.Setup(colorcount,cptr);

  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("arrow,quantity"),
		   TCL_GLOBAL_ONLY);
  if(cptr!=NULL && *cptr!='\0')
    arrow_config.colorquantity=cptr;

  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("arrow,subsample"),
		   TCL_GLOBAL_ONLY);
  if(cptr!=NULL && *cptr!='\0')
    arrow_config.subsample=Nb_Atof(cptr);

  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("arrow,mag"),
		   TCL_GLOBAL_ONLY);
  if(cptr!=NULL && *cptr!='\0')
    arrow_config.mag=Nb_Atof(cptr);



  // Pixel plot configuration
  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("pixel,status"),
		   TCL_GLOBAL_ONLY);
  if(cptr!=NULL && *cptr!='\0')
    pixel_config.displaystate=atoi(cptr);

  colorcount=-1;
  cptr=Tcl_GetVar2(interp,ab1("plot_config"),
		   ab2("pixel,colorcount"),TCL_GLOBAL_ONLY);
  if(cptr!=NULL && *cptr!='\0') colorcount=atoi(cptr);
  if(colorcount<0) colorcount=pixel_config.colormap.GetColorCount();
  cptr=Tcl_GetVar2(interp,ab1("plot_config"),
			 ab2("pixel,colormap"),TCL_GLOBAL_ONLY);
  if(cptr!=NULL && *cptr!='\0')
    pixel_config.colormap.Setup(colorcount,cptr);

  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("pixel,quantity"),
		   TCL_GLOBAL_ONLY);
  if(cptr!=NULL && *cptr!='\0')
    pixel_config.colorquantity=cptr;

  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("pixel,subsample"),
		   TCL_GLOBAL_ONLY);
  if(cptr!=NULL && *cptr!='\0')
    pixel_config.subsample=Nb_Atof(cptr);

  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("pixel,mag"),
		   TCL_GLOBAL_ONLY);
  if(cptr!=NULL && *cptr!='\0')
    pixel_config.mag=Nb_Atof(cptr);

  // Misc.
  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("misc,drawboundary"),
		   TCL_GLOBAL_ONLY);
  if(cptr!=NULL && *cptr!='\0')
    myFrame.SetDrawBoundary(atoi(cptr));

  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("misc,background"),
		   TCL_GLOBAL_ONLY);
  if(cptr!=NULL && *cptr!='\0')
    myFrame.SetBackgroundColor(cptr);

  // Get and set quantity type list
  Nb_List<Nb_DString> cqtypes;
  int cqtypec=myFrame.GetColorQuantityTypes(cqtypes);
  char **cqtypev=new char*[cqtypec];
  int i; Nb_DString* sp;
  for(i=0, sp=cqtypes.GetFirst() ;
      sp!=NULL && i<cqtypec ; i++,sp=cqtypes.GetNext()) {
    cqtypev[i]=new char[sp->Length()+1];
    strcpy(cqtypev[i],sp->GetStr());
  }
  char *cqlist=Tcl_Merge(cqtypec,cqtypev);
  Tcl_SetVar2(interp,ab1("plot_config"),ab2("quantitylist"),
			 cqlist,TCL_GLOBAL_ONLY);
  Tcl_Free(cqlist);
  for(i=0;i<cqtypec;i++) delete[] cqtypev[i];
  delete[] cqtypev;

  // Try to make selected color quantities valid
  int acq_match=0,pcq_match=0;
  const char* dfcq=NULL;
  for(sp=cqtypes.GetFirst() ; sp!=NULL ; sp=cqtypes.GetNext()) {
    if(strcmp(arrow_config.colorquantity,sp->GetStr())==0) acq_match=1;
    if(strcmp(pixel_config.colorquantity,sp->GetStr())==0) pcq_match=1;
    if(dfcq==NULL) dfcq=sp->GetStr();
  }
  if(!acq_match && dfcq!=NULL) {
    arrow_config.colorquantity=dfcq;
    Tcl_SetVar2(interp,ab1("plot_config"),ab2("arrow,quantity"),
		Oc_AutoBuf(dfcq),TCL_GLOBAL_ONLY);
  }
  if(!pcq_match && dfcq!=NULL) {
    pixel_config.colorquantity=dfcq;
    Tcl_SetVar2(interp,ab1("plot_config"),ab2("pixel,quantity"),
		Oc_AutoBuf(dfcq),TCL_GLOBAL_ONLY);
  }

  // Apply new configuration
  myFrame.SetPlotConfiguration(arrow_config,pixel_config);
  return TCL_OK;
}

int SetFrameRotation(ClientData,Tcl_Interp *interp,
		     int argc,char** argv)
{
  Tcl_ResetResult(interp);
  if(argc!=2) {
    sprintf(buf,"SetFrameRotation must be called with"
	    " 1 argument: new_rotation_angle"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  myFrame.SetCoordinates(AngleToCoords(Nb_Atof(argv[1])));

  return TCL_OK;
}

int GetFrameBox(ClientData,Tcl_Interp *interp,
		 int argc,char**)
{ // NOTE: The return value from this routine is a *6*-tuple,
  //  "xmin ymin zmin xmax ymax zmax".
  Tcl_ResetResult(interp);
  if(argc!=1) {
    sprintf(buf,"GetFrameBox should be called with no arguments."
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  Nb_BoundingBox<REAL4> bb=myFrame.GetDisplayBox();
  Nb_Vec3<REAL4> minpt,maxpt;
  bb.GetExtremes(minpt,maxpt);
  if(bb.IsEmpty()) maxpt=minpt;  // Empty boxes may have maxpt<minpt.

  sprintf(buf,"%g %g %g %g %g %g",
	  minpt.x,minpt.y,minpt.z,maxpt.x,maxpt.y,maxpt.z);
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int GetFrameScale(ClientData,Tcl_Interp *interp,
		  int argc,char** argv)
{
  Tcl_ResetResult(interp);
  if(argc!=2) {
    sprintf(buf,"GetFrameScale must be called with"
	    " 1 argument: name of Tk's FrameScale variable"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  sprintf(buf,"%.6g",myFrame.GetZoom());
  Tcl_SetVar(interp,argv[1],buf,0);

  return TCL_OK;
}

int SetFrameScale(ClientData,Tcl_Interp *interp,
		  int argc,char** argv)
{
  Tcl_ResetResult(interp);
  if(argc<2 || argc>3) {
    sprintf(buf,"SetFrameScale must be called with"
	    " either 1 or 2 arguments: <new_scale|new_width new_height>"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  // Change scale in display list
  REAL4m newzoom;
  if(argc==2) newzoom=myFrame.SetZoom(Nb_Atof(argv[1]));
  else        newzoom=myFrame.SetZoom(Nb_Atof(argv[1]),Nb_Atof(argv[2]));

  sprintf(buf,"%g",newzoom);
  Tcl_SetResult(interp,buf,TCL_VOLATILE);

  return TCL_OK;
}

int WriteMesh(ClientData,Tcl_Interp *interp,
	      int argc,char** argv)
{
  Tcl_ResetResult(interp);
  if(argc<4 || argc>6) {
    sprintf(buf,"WriteMesh must be called with"
	    " 3-5 arguments: filename <ascii|binary4|binary8>"
	    " <regular|irregular> [title] [description]"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  int errcode=TCL_OK;

  Vf_OvfDataStyle ods=vf_oascii;
  if(strcmp("binary4",argv[2])==0)      ods=vf_obin4;
  else if(strcmp("binary8",argv[2])==0) ods=vf_obin8;

  BOOL force_irreg=0;
  if(strcmp("irregular",argv[3])==0) force_irreg=1;

  const char *title="",*desc="";
  if(argc>4) title=argv[4];
  if(argc>5) desc=argv[5];

  // Open file
  FILE *fptr=Nb_FOpen(argv[1],"wb");
  if(fptr==NULL) {
    sprintf(buf,"Error in WriteMesh: Unable to open file %s for output",
	    argv[1]);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    errcode=TCL_ERROR;
  }
  else {
    Vf_OvfFileOutput ofo;
    if(ofo.WriteMesh(myMesh,fptr,ods,force_irreg,title,desc)!=0) {
      sprintf(buf,"WriteMesh error");
      Tcl_AppendResult(interp,buf,(char *)NULL);
      errcode=TCL_ERROR;
    }
    fclose(fptr);
  }

  return errcode;
}

int Mmdisp_Init(Tcl_Interp *interp)
{
  // Initialize C++ data structures
  myFrame.SetTclInterp(interp);
  myFrame.SetMesh(myMesh);

  // Register C/C++ routines with interpreter
  Oc_RegisterCommand(interp,"ChangeMesh",ChangeMesh);
  Oc_RegisterCommand(interp,"DrawFrame",DrawFrame);
  Oc_RegisterCommand(interp,"GetFirstFileName",GetFirstFileName);
  Oc_RegisterCommand(interp,"GetFrameBox",GetFrameBox);
  Oc_RegisterCommand(interp,"GetFrameRotation",GetFrameRotation);
  Oc_RegisterCommand(interp,"GetFrameScale",GetFrameScale);
  Oc_RegisterCommand(interp,"GetMeshDescription",GetMeshDescription);
  Oc_RegisterCommand(interp,"GetMeshIncrement",GetMeshIncrement);
  Oc_RegisterCommand(interp,"GetMeshName",GetMeshName);
  Oc_RegisterCommand(interp,"GetMeshTitle",GetMeshTitle);
  Oc_RegisterCommand(interp,"GetMeshType",GetMeshType);
  Oc_RegisterCommand(interp,"GetVecColor",GetVecColor);
  Oc_RegisterCommand(interp,"SetFrameRotation",SetFrameRotation);
  Oc_RegisterCommand(interp,"SetFrameScale",SetFrameScale);
  Oc_RegisterCommand(interp,"SetMeshTitle",SetMeshTitle);
  Oc_RegisterCommand(interp,"UpdatePlotConfiguration",
                     UpdatePlotConfiguration);
  Oc_RegisterCommand(interp,"WriteMesh",WriteMesh);

  // Pass the ServerFlag value to the Tcl interpreter
  Oc_AutoBuf ab1,ab2,ab3;
  sprintf(buf,"%d",ServerFlag);
  Tcl_SetVar(interp,ab1("mmd_serverflag"),buf,0);
    
  // The remnants of the oldest code for loading an initializing the
  // OOMMF extensions.  Should be phased out.
  // Define the standard procs and global variables of the package
  //    The first few from macros...
  Tcl_SetVar2(interp,ab1("_omfmmd_g"),ab2("patchLevel"),
	      ab3(MMDISP_PATCH_LEVEL), TCL_GLOBAL_ONLY);
  Tcl_SetVar2(interp,ab1("_omfmmd_g"),ab2("version"),
	      ab3(MMDISP_VERSION), TCL_GLOBAL_ONLY);
  Tcl_PkgProvide(interp,ab1("Mmdisp"),ab2(MMDISP_VERSION));

    // Evaluates Mmdisp's initialization script
  strncpy(buf2,ab1("proc OMFMMD_PackageName {} {return Mmdisp}"),sizeof(buf2));
  if(Tcl_GlobalEval(interp,buf2) != TCL_OK) return TCL_ERROR;
  strncpy(buf2,ab1("proc OMFMMD_FileNameBase {} {string tolower [OMFMMD_PackageName]}"),sizeof(buf2));
  if(Tcl_GlobalEval(interp,buf2) != TCL_OK) return TCL_ERROR;
  strncpy(buf2,ab1("proc OMFMMD_Version {} [list return $_omfmmd_g(version)]"),sizeof(buf2));
  if(Tcl_GlobalEval(interp,buf2) != TCL_OK) return TCL_ERROR;

    // Register cleanup routine
#if ((TCL_MAJOR_VERSION == 7) && (TCL_MINOR_VERSION >= 5)) \
  || (TCL_MAJOR_VERSION > 7)
       // Note: The following test would break against version number
       //   7.10, but Tcl is going from 7.6 to 8.0, so this shouldn't
       //   be a problem.
       if(Nb_Atof(Tcl_GetVar(interp,ab1("tcl_version"),
			      TCL_GLOBAL_ONLY))>7.4)
	 Tcl_CreateExitHandler(TclCommandsCleanup,0);
  // We check dynamically so that dynamic builds against 7.5+
  //  libraries still have a chance to run using 7.4 libraries.
#endif

  return TCL_OK;
}
