# FILE: host.tcl
#
# An object which plays the role of a host within an interpreter.
#
# The Tcl event loop must be running for this event-driven object to function.
#

Oc_Class Net_Host {

    # Index of host objects by name -- used to avoid duplicate objects.
    array common index {}
    common dir

    ClassConstructor {
        set dir [file dirname [info script]]
    }

    # The port on any host to which this object sends its commands
    public common port 15136

    # Milliseconds to wait for host thread to start on localhost
    public variable startWait = 60000
    # The ID of the current timeout (returned from 'after' command)
    private variable timeoutevent

    # The name of the host this object acts as a proxy for
    const public variable hostname = localhost
    
    # Log for logging messages
    public variable log {
        if {[catch {$log Append "Logging messages generated by $this."}]} {
            error "$log is not able to log messages."
        }
        if {[info exists logHandler]} {
            $logHandler delete
        }
        set logHandler [em_eventHandler new $log delete \
                [list $this FatalError "Log dying"]]
        if {$imadelog} {
            $oldlog Delete
        }
        set oldlog $log
    }
    private variable oldlog = Oc_Nop
    private variable logHandler
    private variable imadelog = 0

    # The connection through which we communicate to the host.
    private variable connection

    private variable msgHandler

    # Socket for receiving ping from localhost OOMMF host server
    private variable listensocket

    # Send id
    private variable id = 0

    # Reply message
    private variable reply

    # Is object ready to relay messages to and from the host?
    private variable ready = 0

    # Event handling support
    private variable bindtags
    private variable myEventHandlers = {}
    private array variable rhandlers = {}
    private array variable toe = {}

    private variable protocolversion

    # Do we display a gui?
    public variable gui = 0 {
	package require Tk
	wm geometry . {} ;# Make sure we have auto-sizing root window
        if {$gui} {
            if {[info exists myWidget]} {
                return
            }
            Net_HostGui New myWidget $parentWin -host $this
            pack [$myWidget Cget -winpath] -side top -fill both -expand 1
        } else {
            if {![info exists myWidget]} {
                return
            }
            $myWidget Delete
            unset myWidget
	    $parentWin configure -width 1 -height 1  ;# Auto-resize
	    $parentWin configure -width 0 -height 0
        }
        upvar #0 ${this}_BtnVar btnvar
        set btnvar $gui
    }
    public variable parentWin = .
    private variable myWidget

    method ToggleGui { pwin } {
        set parentWin $pwin
        upvar #0 ${this}_BtnVar btnvar
        $this Configure -gui $btnvar
    }

    # Constructor is a non-blocking operation.  After it returns an instance
    # id, the Ready method of that instance will return a boolean indicating
    # whether the object is ready.  If not, then when it becomes ready, it
    # will generate an event with id 'Ready'.  If the object cannot become
    # ready, it will delete itself, generating an event with is 'delete'.
    # Users of instances of this class should set up to handle those two
    # events, when the Ready method returns false.
    #
    Constructor { args } {
        eval $this Configure $args
        # Enforce only one object per host
        if {![catch {set index($hostname)} existinghost] \
                && [llength [info commands $existinghost]]} {
            unset hostname
            $this Delete
            array set argarray $args
            catch {unset argarray(-hostname)}
            eval $existinghost Configure [array get argarry]
            return $existinghost
        }
        array set index [list $hostname $this]
        if {![info exists log]} {
            # Provide a default log
            $this Configure -log [em_log New _ -title "Host $hostname Log"]
            set imadelog 1
        }
        # Enable event generation
        set bindtags [list $this $class]
        after 0 $this Connect
    }

    method Connect {args} {
        if {[llength $args]} {
            $log Append "Received ping!"
        }
        if {[catch {
                Net_Connection New connection -hostname $hostname -port $port \
                        -log $log
                } msg]} {
            $this ConnectionFail $msg
        } else {
            $log Append "Net_Host Connect set connection to $connection"
            lappend myEventHandlers [em_eventHandler new $connection delete \
                    [list $this ConnectionFail]]
            lappend myEventHandlers [em_eventHandler new $connection Ready \
                    [list $this ConnectionReady]]
        }
        # NOTE: It is important that the ping channel be closed *after*
        # the new Net_Connection instance is created.  Otherwise Tcl may
        # try to re-use the same socket name, and if there are any bugs
        # in the Tcl library code which cleans up closed sockets (as it
        # appears there are, at least on some platforms), then there is the
        # danger of "cross-talk" between the closed ping socket and the new
        # connection socket.  In particular, we have observed spurious EOF
        # conditions appearing on the new Net_Connection's socket when the
        # order of these operations was reversed in an earlier version of this
        # software.  Better to avoid trouble by forcing Tcl to give us a 
        # socket with a new name.
        if {[llength $args]} {
            catch {$log Append "Closing ping channel!"}
            close [lindex $args 0]
        }
    }

    method ConnectionFail {{why ""}} {
        if {$ready} {
            $this FatalError "Connection to $hostname died.\n$why"
            return
        }
        if {[string match localhost $hostname] && ![info exists listensocket]} {
            if {[info exists connection]} {
                unset connection
            }
            $this StartLocalHost
        } else {
            $this FatalError "Can't open connection to $hostname.\n$why"
        }
    }

    method ConnectionReady {} {
        set msgHandler [em_eventHandler new $connection Readable \
                [list $this VerifyOOMMFHost]]
    }

    method VerifyOOMMFHost {} {
        if {[info exists listensocket]} {
            close $listensocket
            unset listensocket
        }
        if {[info exists timeoutevent]} {
            after cancel $timeoutevent
            unset timeoutevent
        }
        $msgHandler delete
        unset msgHandler
        if {![regexp {OOMMF host protocol ([0-9.]+)} [$connection Get] \
                match version]} {
            # Not an OOMMF host
            $this FatalError "No OOMMF host server answering on $hostname"
            return
        }
        set protocolversion $version
        $log Append "OOMMF host $hostname ready"
        set ready 1
        em_eventHandler Generate Ready $bindtags
    }

    method ProtocolVersion {} {
        if {!$ready} {
            error "host $hostname not ready"
        }
        return $protocolversion
    }

    # Launch an OOMMF host server on localhost
    private method StartLocalHost {} {
        set hostthread [file join $dir threads host.tcl]
        if {[file readable $hostthread]} {
            # Set up to receive ping from host thread or timeout
            set listensocket [socket -server [list $this Connect] \
		    -myaddr localhost 0]  ;# Force to localhost for now
            set listenport [lindex [fconfigure $listensocket -sockname] 2]
            set timeoutevent [after $startWait $this StartTimeout]
            $log Append "Starting OOMMF host server on localhost"
            # Ought to redirect std channels somewhere
            exec [info nameofexecutable] $hostthread -notk $port $listenport > \
                    [[Oc_Config RunPlatform] GetValue path_device_null] &
        } else {
            error "Installation error: No OOMMF host server available to start"
        }
    }

    method StartTimeout {} {
        $this FatalError \
                "Timed out waiting for OOMMF host server to start on localhost"
    }

    method Receive { rid } {  
        $rhandlers($rid) delete
        unset rhandlers($rid)
        $toe($rid) delete
        unset toe($rid)
        set reply [$connection Get] 
        em_eventHandler Generate Reply$rid $bindtags
    }

    method Timeout { rid } {
        $rhandlers($rid) delete
        unset rhandlers($rid)
        $toe($rid) delete
        unset toe($rid)
        $log Append "Timed out waiting for host $hostname reply to query $rid"
        em_eventHandler Generate Timeout$rid $bindtags
    }

    method Get {} {
        return $reply
    }

    # Send a command to the host
    method Send { args } {
        if {!$ready} {
            error "host $hostname not ready"
        }
        incr id
        set qid [eval $connection Query $args]
        array set rhandlers [list $id [em_eventHandler new $connection \
                Reply$qid [list $this Receive $id]]]
        array set toe [list $id [em_eventHandler new $connection Timeout$qid \
                [list $this Timeout $id]]]
        return $id
    }

    method Ready {} {
        return $ready
    }

    method FatalError { msg } {
        catch {$log Append $msg}
        catch {Ow_Dialog 1 "Error in $class instance $this:" error $msg \
                {} 0}
        $this Delete
    }

    Destructor {
        set ready 0
        if {[info exists bindtags]} {
            em_eventHandler Generate delete $bindtags
        }
        # delete handlers before deleting event-generating members
        if {[info exists myEventHandlers]} {
        foreach handler $myEventHandlers {
            $handler delete
        }
        }
        foreach idx [array names rhandlers] {
            $rhandlers($idx) delete
        }
        foreach idx [array names toe] {
            $toe($idx) delete
        }
        if {[info exists msgHandler]} {
            $msgHandler delete
        }
        if {[info exists connection]} {
            $log Append "Net_Host delete destroying connection $connection"
            $connection Delete
        }
        if {[info exists timeoutevent]} {
            after cancel $timeoutevent
        }
        if {[info exists listensocket]} {
            close $listensocket
        }
        if {[info exists log]} {
            $log Append "Deleting $this ..."
            $this Configure -log Oc_Nop
        }
        if {[info exists myWidget]} {
            $this Configure -gui 0
        }
        if {[info exists hostname]} {
            unset index($hostname)
        }
    }

}
